VERSION_NAME = "tmtccmd"
VERSION_MAJOR = 1
VERSION_MINOR = 10
VERSION_REVISION = 1

# I think this needs to be in string representation to be parsed so we can't
# use a formatted string here.
__version__ = "1.10.1"
